# get from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52275
#!/usr/bin/env python
#
# 
# a python sparse matrix class
# A. Pletzer 5 Jan 00
#
from types import *
from math import *
import vector

class sparse:
	"""
	A sparse matrix class supporting elementwise operations, dot products,
	and conjugate gradient solver.
	A. Pletzer 5 Jan 00

	Dictionary storage format { (i1,j1): value1, (i2,j2): value2, ... }
	where (i,j) are the matrix indices
	
	"""

	def __init__(self, start={}):
		if type(start) == DictType:
			self.data = start
		elif isSparse(start):
			self = start
		else:
			raise TypeError, self.__class__,'in __init__'

	def __del__(self):
		del self.data

	def __len__(self):
		return len(self.data)

	def size(self):
		nrow = 0
		ncol = 0
		for key in self.data.keys():
			nrow = max([nrow, key[0]+1])
			ncol = max([ncol, key[1]+1])
		return [nrow, ncol]

	def __getitem__(self, elem):
		if type(elem) != TupleType or len(elem) != 2:
			print '**__getitem__ argument error'
		else:
			return self.data[elem]

	def __setitem__(self, elem, value):
		self.data[elem] = value

	def __delitem__(self, elem):
		del self.data[elem]

	def __add__(self, other):
		new = sparse({})
		if not isSparse(other):
			raise TypeError, self.__class__,'in __add__'
		else:
			for k in self.data.keys():
				new.data[k] = self.data[k]
			for k in other.data.keys():
				try:
					new.data[k] = new.data[k] + other.data[k]
				except:
					new.data[k] = other.data[k]
		return new
		
	def __neg__(self):
		new = sparse({})
		for k in self.data.keys():
			new[k] = -self[k]
		return new

	def __sub__(self, other):
		new = sparse({})
		if not isSparse(other):
			raise TypeError, self.__class__,'in __sub__'
		else:
			for k in self.data.keys():
				new.data[k] = self.data[k]
			for k in other.data.keys():
				try:
					new.data[k] = new.data[k] - other.data[k]
				except:
					new.data[k] = -other.data[k]
		return new


	def __mul__(self, other):
		# element by element multiplication
		if isSparse(other):
			new = sparse({})
			for k in self.data.keys():
				new.data[k] = self.data[k]*other.data.get(k,0.)
			return new
		elif type(other)==FloatType or type(other)==IntType or  type(other)==ComplexType:
			new = sparse({})
			for k in self.data.keys():
				new.data[k] = self.data[k]*other
			return new
		else:
			raise TypeError, self.__class__,'in __mul__'


	def __rmul__(self, other):
		if type(other)==FloatType or type(other)==IntType or  type(other)==ComplexType:
			new = sparse({})
			for k in self.data.keys():
				new.data[k] = self.data[k] * other
			return new
		else:
			raise TypeError, self.__class__,'in __mul__'


	def __div__(self, other):
		if type(other)==FloatType or type(other)==IntType or  type(other)==ComplexType:
			new = sparse({})
			for k in self.data.keys():
				new.data[k] = self.data[k] / other
			return new
		else:
			raise TypeError, self.__class__,'in __div__'



	def out(self):
		print '# (i, j) -- value'
		for k in self.data.keys():
			print k, self.data[k]

	def plot(self, width_in=400, height_in=400):

		try:
			from Tkinter import *
			import colormap


			cmax = max(self.data.values())
			cmin = min(self.data.values())

			offset =  0.05*min(width_in, height_in)
			xmin, ymin, xmax, ymax = 0,0,self.size()[0], self.size()[1]
			scale =  min(0.9*width_in, 0.9*height_in)/max(xmax-xmin, ymax-ymin)

			root = Tk()
			frame = Frame(root)
			frame.pack()

			text = Label(width=20, height=10, text='matrix sparsity')
			text.pack()


			canvas = Canvas(bg="white", width=width_in, height=height_in)
			canvas.pack()

			button = Button(frame, text="OK?", fg="red", command=frame.quit)
			button.pack()

			for index in self.data.keys():
				ix, iy = index[0], ymax-index[1]-1
				xa, ya = offset+scale*(ix  ), height_in -offset-scale*(iy  )
				xb, yb = offset+scale*(ix+1), height_in -offset-scale*(iy  )
				xc, yc = offset+scale*(ix+1), height_in -offset-scale*(iy+1)
				xd, yd = offset+scale*(ix  ), height_in -offset-scale*(iy+1)
				color = colormap.strRgb(self.data[index], cmin, cmax)
				canvas.create_polygon(xa, ya, xb, yb, xc, yc, xd, yd, fill=color)

			root.mainloop()

		except: print 'ERROR ',self.__class__,'::plot Missing modules?'

	def CGsolve(self, x0, b, tol=1.0e-10, nmax = 1000, verbose=1):
		"""
		Solve self*x = b and return x using the conjugate gradient method
		"""
		try:
			if not vector.isVector(b):
				raise TypeError, self.__class__,' in solve '
			else:
				if self.size()[0] != len(b) or self.size()[1] != len(b):
					print '**Incompatible sizes in solve'
					print '**size()=', self.size()[0], self.size()[1]
					print '**len=', len(b)
				else:
					kvec = diag(self) # preconditionner
					n = len(b)
					x = x0 # initial guess
					r =  b - dot(self, x)
					try:
						w = r/kvec
					except: print '***singular kvec'
					p = vector.zeros(n);
					beta = 0.0;
					rho = vector.dot(r, w);
					err = vector.norm(dot(self,x) - b);
					k = 0
					if verbose: print " conjugate gradient convergence (log error)"
					while abs(err) > tol and k < nmax:
						p = w + beta*p;
						z = dot(self, p);
						alpha = rho/vector.dot(p, z);
						r = r - alpha*z;
						w = r/kvec;
						rhoold = rho;
						rho = vector.dot(r, w);
						x = x + alpha*p;
						beta = rho/rhoold;
						err = vector.norm(dot(self, x) - b);
						if verbose: print k,' %5.1f ' % log10(err)
						k = k+1
					return x
				
		except: print 'ERROR ',self.__class__,'::CGsolve '
		
	    		
	def biCGsolve(self,x0, b, tol=1.0e-10, nmax = 1000):
		
		"""
		Solve self*x = b and return x using the bi-conjugate gradient method
		"""

		try:
			if not vector.isVector(b):
				raise TypeError, self.__class__,' in solve '
			else:
				if self.size()[0] != len(b) or self.size()[1] != len(b):
					print '**Incompatible sizes in solve'
					print '**size()=', self.size()[0], self.size()[1]
					print '**len=', len(b)
				else:
					kvec = diag(self) # preconditionner 
					n = len(b)
					x = x0 # initial guess
					r =  b - dot(self, x)
					rbar =  r
					w = r/kvec;
					wbar = rbar/kvec;
					p = vector.zeros(n);
					pbar = vector.zeros(n);
					beta = 0.0;
					rho = vector.dot(rbar, w);
					err = vector.norm(dot(self,x) - b);
					k = 0
					print " bi-conjugate gradient convergence (log error)"
					while abs(err) > tol and k < nmax:
						p = w + beta*p;
						pbar = wbar + beta*pbar;
						z = dot(self, p);
						alpha = rho/vector.dot(pbar, z);
						r = r - alpha*z;
						rbar = rbar - alpha* dot(pbar, self);
						w = r/kvec;
						wbar = rbar/kvec;
						rhoold = rho;
						rho = vector.dot(rbar, w);
						x = x + alpha*p;
						beta = rho/rhoold;
						err = vector.norm(dot(self, x) - b);
						print k,' %5.1f ' % log10(err)
						k = k+1
					return x
			
		except: print 'ERROR ',self.__class__,'::biCGsolve'
			

###############################################################################

def isSparse(x):
    return hasattr(x,'__class__') and x.__class__ is sparse

def transp(a):
	"""
	Return transpose
	"""
	if isSparse(a):
		new = sparse({})
		for key in a.data.keys():
			new.data[(key[1], key[0])] = a.data[key]
		return new
	else:
		raise TypeError, ' in transp(sparse)'


def dot(a, b):
	"""
	Dot or inner product
	"""

	try:
		if isSparse(a) and vector.isVector(b):
			new = vector.zeros(a.size()[0])
			for ij in a.data.keys():
				i=ij[0]
				j=ij[1]
				new[i] = new[i] + a.data[ij]* b[j]
			return new
		elif vector.isVector(a) and isSparse(b):
			new = vector.zeros(b.size()[1])
			for ij in b.data.keys():
				i=ij[0]
				j=ij[1]
				new[j] = new[j] + a[i]* b.data[ij]
			return new
		elif isSparse(a) and isSparse(b):
			if a.size()[1] != b.size()[0]:
				print '**Warning shapes do not match in dot(sparse, sparse)'
			new = sparse({})
			n = min([a.size()[1], b.size()[0]])
			for i in range(a.size()[0]):
				for j in range(b.size()[1]):
					sum = 0
					for k in range(n):
						sum = sum + a.data.get((i,k),0.)*b.data.get((k,j),0.)
					if sum != 0:
						new[(i,j)] = sum
			return new
		else:
			raise TypeError, 'in dot'

	except: print 'ERROR sparse::dot'

def diag(b):
	"""
	Given a sparse matrix b return its diagonal as a vector
	"""

	try:
		res = vector.zeros(b.size()[0])
		for i in range(b.size()[0]):
			try:
				res[i] = b[(i,i)]
			except:
				res[i] = 0.0
		return res
	
	except: print 'ERROR sparse::diag'
		
def identity(n):
	"""
	Returns the identity sparse matrix
	"""
	if type(n) != IntType:
		raise TypeError, ' in identity: # must be integer'
	else:
		new = sparse({})
		for i in range(n):
			new.data[(i,i)] = 1
		return new

###############################################################################
if __name__ == "__main__":

	print 'a = sparse()'
	a = sparse()

	print 'a.__doc__=',a.__doc__

	print 'a[(0,0)] = 1.0'
	a[(0,0)] = 1.0
	a.out()

	print 'a[(2,3)] = 3.0'
	a[(2,3)] = 3.0
	a.out()

	print 'len(a)=',len(a)
	print 'a.size()=', a.size()
			
	b = sparse({(0,0):2.0, (0,1):1.0, (1,0):1.0, (1,1):2.0, (1,2):1.0, (2,1):1.0, (2,2):2.0})
	print 'a=', a.data
	print 'b=', b.data
	b.out()

	print 'a+b'
	c = a + b
	c.out()

	print '-a'
	c = -a
	c.out()
	a.out()

	print 'a-b'
	c = a - b
	c.out()

	print 'a*1.2'
	c = a*1.2
	c.out()


	print '1.2*a'
	c = 1.2*a
	c.out()
	print 'a=', a.data

	print 'dot(a, b)'
	print 'a.size()[1]=',a.size()[1],' b.size()[0]=', b.size()[0]
	c = dot(a, b)
	c.out()

	print 'dot(b, a)'
	print 'b.size()[1]=',b.size()[1],' a.size()[0]=', a.size()[0]
	c = dot(b, a)
	c.out()

	try:
		print 'dot(b, vector.vector([1,2,3]))'
		c = dot(b, vector.vector([1,2,3]))
		c.out()
	
		print 'dot(vector.vector([1,2,3]), b)'
		c = dot(vector.vector([1,2,3]), b)
		c.out()

		print 'b.size()=', b.size()
	except: pass
	
	print 'a*b -> element by element product'
	c = a*b
	c.out()

	print 'b*a -> element by element product'
	c = b*a
	c.out()
	
	print 'a/1.2'
	c = a/1.2
	c.out()

	print 'c = identity(4)'
	c = identity(4)
	c.out()

	print 'c = transp(a)'
	c = transp(a)
	c.out()


	b[(2,2)]=-10.0
	b[(2,0)]=+10.0

	try:
		import vector
		print 'Check conjugate gradient solver'
		s = vector.vector([1, 0, 0])
		print 's'
		s.out()
		x0 = s 
		print 'x = b.biCGsolve(x0, s, 1.0e-10, len(b)+1)'
		x = b.biCGsolve(x0, s, 1.0e-10,  len(b)+1)
		x.out()

		print 'check validity of CG'
		c = dot(b, x) - s
		c.out()
	except: pass

	print 'plot b matrix'
	b.out()
	b.plot()

	print 'del b[(2,2)]'
	del b[(2,2)]


