# get from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52272
#!/usr/bin/env python
#
# 
# a python vector class
# A. Pletzer 5 Jan 00
#
from types import *
import math

"""
A vector class that supports elementwise mathematical operations
(sin, cosh, etc) and other common external operations on vectors
such as 'dot' for the inner product, 'sum', 'norm' etc.
"""

class vector:
	"""
        A vector class supporting elementwise operations
	"""

	def __init__(self, start=[]):
	    """
	    Initializes the list data structure.
	    Default is an empty list.
	    """
	    if type(start) == ListType:
		    self.data = start
	    elif isVector(start):
		    self = start
	    else:
		raise TypeError, self.__class__,'in __init__'

	def __del__(self):
	    """
	    Overridden deletion function.
	    """
	    del self.data

	def __len__(self):
	    """
	    Overridden len function.
	    """
	    return len(self.data)

	def size(self):	   
            """
	    Returns the size of the vector.
	    """
	    return len(self)

	def __getitem__(self, elem):
	    """
	    Overridden getitem operator(no error checking).
	    Returns the index specified.
	    """
	    return self.data[elem]

	def __setitem__(self, elem, value):
	    """
	    Overridden setitem operator(accomodates any positive index).
	    Usage: <vector object>[<index>] = x
	    """
	    try:
		self.data[elem] = value
	    except:
		# increase vector length
		self.data.append(0.0)
		self[elem] = value

	def __getslice__(self, i, j):
	    """
	    Overridden getslice operator.
	    """
	    return vector(self.data[i:j])

	def __setslice__(self, i, j, list):
	    """
	    Overridden setslice operator.
	    """
	    if type(list) != ListType:
		raise TypeError, self.__class__,'in __setslice__'
	    for index in xrange(i, j):
		try:
		    self[index] = list[index-i]
		except:
		    print '**Index out of range'

	def __add__(self, other):
	    """
	    Addition operator.
	    """
	    new = vector([])
	    if not isVector(other):
		    raise TypeError, self.__class__,'in __add__'
	    else:
		    if len(self) != len(other):
			    print '**Lengths do not match'
		    else:
			    for i in xrange(0, len(self)):
				    new[i] = self[i] + other[i]
	    return new
		
	def __neg__(self):
	    """
	    Minus operator.
	    """
	    new = vector([])
	    for i in xrange(0, len(self)):
		    new[i] = -self[i]
	    return new

	def __sub__(self, other):
	    """
	    Subtraction operator.
	    """
	    new = vector([])
	    if not isVector(other):
		    raise TypeError, self.__class__,'in __add__'
	    else:
		    if len(self) != len(other):
			    print '**Lengths do not match'
		    else:
			    for i in xrange(0, len(self)):
				    new[i] = self[i] - other[i]
	    return new

	def __mul__(self, other):
	    """
	    Multiplication operator,element by element.
	    """
	    # element by element multiplication
	    if isVector(other):
		    if len(self) != len(other):
			    print '**Lengths do not match'
		    else:
			    new = vector([])
			    for i in xrange(0, len(self)):
				    new[i] = self[i]*other[i]
		    return new
	    elif type(other)==FloatType or type(other)==IntType or  type(other)==ComplexType:
		    new = vector([])
		    for i in xrange(0, len(self)):
			    new[i] = self[i] * other
		    return new
	    else:
		raise TypeError, self.__class__,'in __mul__'


	def __rmul__(self, other):
	    """
	    Multiplication operator taking arguments other then 
            vector objects.
	    """
	    if type(other)==FloatType or type(other)==IntType or  type(other)==ComplexType:
		    new = vector([])
		    for i in xrange(0, len(self)):
			    new[i] = self[i] * other
		    return new
	    else:
		    raise TypeError, self.__class__,'in __mul__'


	def __div__(self, other):
	    """
	    Division operator. Divides each element of the vector by the 
	    second argument.
	    """
	    new = vector([])
	    n = len(self)
	    if type(other)==FloatType or type(other)==IntType or  type(other)==ComplexType:
		    for i in xrange(0, n):
			    new[i] = self[i] / other
	    else:
		    for i in xrange(0, n):
			    new[i] = self[i] / other[i]
	    return new

	def __rdiv__(self, other):
	    """
	    The same as __div__
	    """
	    new = vector([])
	    n = len(self)
	    if type(other)==FloatType or type(other)==IntType or  type(other)==ComplexType:
		    for i in xrange(0, n):
			    new[i] = other[i] / self[i]
	    else:
		    for i in xrange(0, n):
			    new[i] = other / self[i]
	    return new


	def out(self):
	    """
	    Prints out the vector.
	    """
	    print self.data	

###############################################################################


def isVector(x):
    """
    Determines if the argument is a vector class object.
    """
    return hasattr(x,'__class__') and x.__class__ is vector

def zeros(n):
    """
    Returns a zero vector of length n.
    """
    new = vector([])
    for i in xrange(0, n):
	new[i] = 0.0
    return new

def ones(n):
    """
    Returns a vector of length n with all ones.
    """
    new = vector([])
    for i in xrange(0, n):
	new[i] = 1.0
    return new

def random(n, lmin=0.0, lmax=1.0):

    """
    Returns a random vector of length n.
    """
    try:
	    import whrandom
    
	    new = vector([])
	    gen = whrandom.whrandom()
	    for i in range(n):
		    new[i] = lmin + (lmax-lmin)*gen.random()
	    return new
    
    except: print 'ERROR::random Could not import whrandom'
	
def dot(a, b):
    """
    Computes the dot product of two vectors.
    """
    if isVector(a) and isVector(b) and len(a)==len(b):
	return reduce(lambda x, y: x+y, a*b, 0)
    else:
	raise TypeError, 'in dot'
	

def norm(a):
    """
    Computes the norm of vector a.
    """
    if isVector(a):
	return math.sqrt(dot(a,a))
    else:
	raise TypeError, 'in norm'

def sum(a):
    """
    Returns the sum of the elements of a.
    """
    if isVector(a):
	return reduce(lambda x, y: x+y, a, 0)
    else:
	raise TypeError, 'in sum'

# elementwise operations
	
def log10(a):
    """
    log10 of each element of a.
    """
    if isVector(a):
	return map(math.log10, a)
    else:
	raise TypeError, 'in log10'

def log(a):
    """
    log of each element of a.
    """
    if isVector(a):
	return map(math.log, a)
    else:
	raise TypeError, 'in log'
	    
def exp(a):
    """
    Elementwise exponential.
    """
    if isVector(a):
	return map(math.exp, a)
    else:
	raise TypeError, 'in exp'

def sin(a):
    """
    Elementwise sine.
    """
    if isVector(a):
	return map(math.sin, a)
    else:
	raise TypeError, 'in sin'
	    
def tan(a):
    """
    Elementwise tangent.
    """
    if isVector(a):
	return map(math.tan, a)
    else:
	raise TypeError, 'in tan'
	    
def cos(a):
    """
    Elementwise cosine.
    """
    if isVector(a):
	return map(math.cos, a)
    else:
	raise TypeError, 'in cos'

def asin(a):
    """
    Elementwise inverse sine.
    """
    if isVector(a):
	return map(math.asin, a)
    else:
	raise TypeError, 'in asin'

def atan(a):
    """
    Elementwise inverse tangent.
    """	
    if isVector(a):
	return map(math.atan, a)
    else:
	raise TypeError, 'in atan'

def acos(a):
    """
    Elementwise inverse cosine.
    """
    if isVector(a):
	return map(math.acos, a)
    else:
	raise TypeError, 'in acos'

def sqrt(a):
    """
    Elementwise sqrt.
    """
    if isVector(a):
	return map(math.sqrt, a)
    else:
	raise TypeError, 'in sqrt'

def sinh(a):
    """
    Elementwise hyperbolic sine.
    """
    if isVector(a):
	return map(math.sinh, a)
    else:
	raise TypeError, 'in sinh'

def tanh(a):
    """
    Elementwise hyperbolic tangent.
    """
    if isVector(a):
	return map(math.tanh, a)
    else:
	raise TypeError, 'in tanh'

def cosh(a):
    """
    Elementwise hyperbolic cosine.
    """
    if isVector(a):
	return map(math.cosh, a)
    else:
	raise TypeError, 'in cosh'


def pow(a,b):
    """
    Takes the elements of a and raises them to the power of the
    corresponding element of b.
    """
    if isVector(a):
	if not isVector(b) or type(b)!=ListType:
	    bv = b*ones(len(a))
	    return map(math.pow, a, bv)
	else:
	    try:
		return map(math.pow, a, b)
	    except:
		raise 'cannot take pow'
    else:
	raise TypeError, 'in pow'
	
def atan2(a,b):    
    """
    
    """
    if isVector(a) and isVector(b) and len(a)==len(b):
	return map(math.atan2, a, b)
    else:
	raise TypeError, 'in atan2'
	

###############################################################################
if __name__ == "__main__":

	print 'a = vector()'
	a = vector()

	print 'a.__doc__=',a.__doc__

	print 'a[0] = 1.0'
	a[0] = 1.0

	print 'a[3] = 3.0'
	a[3] = 3.0

	print 'a[0]=', a[0]
	print 'a[1]=', a[1]

	print 'len(a)=',len(a)
	print 'a.size()=', a.size()
			
	b = vector([1, 2, 3, 4])
	print 'a=', a.data
	print 'b=', b.data

	print 'a+b'
	c = a + b
	c.out()

	print '-a'
	c = -a
	c.out()
	a.out()

	print 'a-b'
	c = a - b
	c.out()

	print 'a*1.2'
	c = a*1.2
	c.out()


	print '1.2*a'
	c = 1.2*a
	c.out()
	print 'a=', a.data

	print 'dot(a,b) = ', dot(a,b)
	print 'dot(b,a) = ', dot(b,a)

	print 'a*b'
	c = a*b
	c.out()
	
	print 'a/1.2'
	c = a/1.2
	c.out()

	print 'a[0:2]'
	c = a[0:2]
	c.out()

	print 'a[2:5] = [9.0, 4.0, 5.0]'
	a[2:5] = [9.0, 4.0, 5.0]
	a.out()

	print 'sqrt(a)=',sqrt(a)
	print 'pow(a, 2*ones(len(a)))=',pow(a, 2*ones(len(a)))
	print 'pow(a, 2)=',pow(a, 2*ones(len(a)))

	print 'ones(10)'
	c = ones(10)
	c.out()

	print 'zeros(10)'
	c = zeros(10)
	c.out()	

	print 'del a'
	del a

	try:
		a = random(11, 0., 2.)
		a.out()

	except: pass

